/*	
 *	clippsid.c
 *
 *	P/ECE PSID Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Sat Dec 27 06:00:00 JST 2003 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *
 ****************************************************************************/

/* [`/t[`̌ĂяovO */
static unsigned char psid_program[] = {
	'\x20',		/* +0: JSR Absolute */
	'\x00',		/* +1: ɌĂяoAhX̉ʃoCgZbg */
	'\x00',		/* +2: ɌĂяoAhX̏ʃoCgZbg */
	'\xcb',		/* +3: WAI */
};			/* =4 */
/* ĂяovO]AhX
 * Kernel ROM area (0xe000`0xffff) ̐^񒆂gȂƉ肵āA0xf000`ɒuƂɂ܂B
 */
#define PSID_PROGRAM_ADDRESS	0xf000 /*  */
/* Ăяo([`/t[`)AhXZbgAhX
 * ĂяovÓuJSR Absolutev߂̃IyhƂȂ܂B
 */
#define PSID_RUN_PC_ADDRESS	(PSID_PROGRAM_ADDRESS + 1)

void
psid_m6502_run(PSIDDRIVER* psid, int a, int pc, int cycle)
{
	M6502* m6502 = &psid->m6502;
	unsigned char* mem = psid->mem;

	/* ĂяovOoRŁAĂяovOĂяo܂B */
	mem[PSID_RUN_PC_ADDRESS + 0] = pc & 0xf;
	mem[PSID_RUN_PC_ADDRESS + 1] = pc >> 8;
	m6502->a = a;
	m6502->pc = PSID_PROGRAM_ADDRESS;
	m6502->sp = 0xff;
	m6502->wait = 0;

	/* 荞ݓȂ̂ŁA0xe00`0xfffj󂷂m6502_run2()͗psł!!
	 * ʏm6502_run()g܂B
	 */
	m6502_run(m6502, cycle);
}

/****************************************************************************
 *	PSIDhCo֐
 ****************************************************************************/

int
psid_init(PSIDDRIVER* psid, const void* _data, int len, int i_song)
{
	M6502* m6502 = &psid->m6502;
	unsigned char* mem = psid->mem;
	SID* sid = &psid->sid;
	//
	unsigned char* data = (unsigned char*)_data;
	PSIDHEADER* header = (PSIDHEADER*)_data;
	int data_offset;
	int load_addr;
	int init_addr;
	int play_addr;
	int songs;
	int start_song;
	//
	int n1, n2;

	/* ܂NA܂B */
	memset(psid, 0, sizeof(PSIDDRIVER));

	/* C܂B */
	memset(&mem[0x0000], 0x00, 0xe000);
	memset(&mem[0xe000], 0x40, 0x2000);	/* BASIC/Kernal ROM <= 0x40:RTI */
	mem[0x0001] = 7;			/* 6510 On-chip I/O Register <= 7 (Bank-sel) */
	mem[0x0314] = 0x31;			/* Hardware IRQ Interrupt Address <= 0xea31 (Default) */
	mem[0x0315] = 0xea;

	/* [`/t[`ĂяovO]܂B */
	memcpy(&mem[PSID_PROGRAM_ADDRESS], psid_program, sizeof psid_program);

	/* headert@Cwb_͂܂B */
	if(memcmp(header->signature, "PSID", 4) != 0) return -1; /* VOl` */
	/* version͎g܂B */
	data_offset = PSID_HALF(header->data_offset);
	load_addr = PSID_HALF(header->load_addr);
	init_addr = PSID_HALF(header->init_addr);
	play_addr = PSID_HALF(header->play_addr);
	songs = PSID_HALF(header->songs);
	start_song = PSID_HALF(header->start_song);
	/* name͎g܂B */
	/* author͎g܂B */
	/* copyright͎g܂B */
	/* flags͎g܂B */
	/* reserved͎g܂B */
	/* speed͎g܂B
	 *	bit0=0:PAL/1:NTSC ... {vOPAL^C~Õt@Cz肵Ă܂񂪁A
	 *			      NTSC^C~Õt@CPAL^C~OŃt܂B
	 *	bit1`31:???      ... Ӗ悤łA܂悭킩Ă܂B
	 */

	/* tȔԍ肵܂B */
	if(!songs) songs = 1;			/* Ȑ0 => Ȑ1Ɖ߂dl */
	if(start_song) start_song--;		/* PSIDHEADER.start_song(1x[X) => init_addr(0x[X) */
	if(i_song < 0) i_song = start_song;	/* ɂȔԍw薳(i_song<0)ȂAstart_songgp */
	if(i_song < 0 || songs - 1 < i_song) return -1;

	/* C[Wǂݍ݂܂B */
	if(!load_addr) {
		/* load_addr=NULL̏ꍇ́AC[W̍ŏɃ[hAhXLĂ܂B */
		load_addr = data[data_offset + 0] | data[data_offset + 1] << 8; /* Little-endian! */
		data_offset += 2;
	}
	n1 = len - data_offset;		/* C[WTCY */
	n2 = 0x10000 - load_addr;	/* [h\ȃTCY */
	memcpy(&mem[load_addr], &data[data_offset], n1 <= n2 ? n1 : n2);

	/* SIDZbgB */
	sid_reset(sid, PSID_C64_SID_CLOCK);

	/* CPUZbgB */
	m6502_reset(m6502);
	/* READM6502RAREADO֐g܂B */
	m6502->write = psid_m6502_write;

	/* [`sB */
	if(!init_addr) {
		/* init_addr=NULL̏ꍇ́AC[W̐擪ɏ[`܂B */
		init_addr = load_addr;
	}
	psid_m6502_run(psid, i_song, init_addr, PSID_C64_CPU_CLOCK);

	/* t[`AhXB */
	if(!play_addr) {
		/* play_addr=NULL̏ꍇ́A[`荞݃xN^ݒ肵Ă܂B */
		if(mem[0x0001] & 2) {	/* Kernal ROM switched in */
			play_addr = mem[0x0314] | mem[0x0315] << 8;
		} else {		/* Kernal ROM switched out */
			play_addr = mem[0xfffe] | mem[0xffff] << 8;
		}
	}
	psid->play_addr = play_addr;

	return 0;
}

int
psid_stream_callback(short* wbuff, int param)
{
	PSIDDRIVER* psid = (PSIDDRIVER*)param;

	psid_m6502_run(psid, 0, psid->play_addr, PSID_C64_CPU_CLOCK / 50);
	m6502_int(&psid->m6502);
	sid_mix(&psid->sid, wbuff);

	return 0;
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

int
psid_play(const void* data, int len, int i_song)
{
	static PSIDDRIVER psid; /* STATICł! */

	/* ܂mɒ~܂B */
	psid_stop();

	/* PSIDhCo܂B */
	if(psid_init(&psid, data, len, i_song) != 0) return -1;

	/* Xg[ĐJn܂B */
	stream_play(PSIDBUFLEN, psid_stream_callback, (int)&psid);

	return 0;
}

void
psid_stop()
{
	/* Xg[Đ~܂B */
	stream_stop();

	/* PSIDhCõN[Abv͕svłB */
}

