/*	
 *	clipsid.c
 *
 *	P/ECE SID(MOS6581) Emulator
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Sat Dec 27 06:00:00 JST 2003 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/****************************************************************************
 *
 ****************************************************************************/

/* mCYWFl[^e[u */
const char sid_noise_table[256] = {
	'\x37','\x4a','\xab','\x8e','\xfe','\x31','\x2d','\x1c','\xc7','\x1e','\xd0','\x47','\x9e','\x4e','\xbd','\xea',
	'\xcd','\x3b','\x62','\x22','\xe2','\xa3','\x6a','\x49','\xa8','\x82','\xc1','\xfd','\x2b','\x12','\x93','\x19',
	'\xb8','\xd1','\x4d','\xba','\xda','\x7a','\x9a','\x3a','\x5d','\x09','\x66','\x37','\x4d','\xba','\xdd','\x8c',
	'\xf7','\x0a','\x6c','\x56','\xe5','\xb4','\xbe','\xef','\xe5','\xb4','\xbf','\xf5','\x03','\x47','\x9c','\x43',
	'\x89','\xe8','\xbf','\xf6','\x05','\x53','\xd6','\x68','\x42','\x84','\xcc','\x37','\x4c','\xb7','\xce','\x3e',
	'\x70','\x69','\x46','\x9a','\x39','\x55','\xe2','\xa5','\x73','\x77','\x8e','\xfd','\x2a','\x0c','\x74','\x7d',
	'\xab','\x8f','\x02','\x41','\x80','\xbc','\xe3','\xa6','\x78','\x8f','\x03','\x47','\x9c','\x44','\x8e','\x00',
	'\x39','\x58','\xf1','\xed','\xdc','\x86','\xd7','\x6b','\x50','\xca','\x2c','\x15','\xa2','\x61','\x1c','\xc7',
	'\x1c','\xc5','\x14','\x9d','\x4c','\xb4','\xbf','\xf4','\xfd','\x2a','\x09','\x67','\x3b','\x61','\x1f','\xd2',
	'\x53','\xd7','\x6b','\x51','\xd0','\x47','\x9b','\x41','\x7e','\xad','\x9a','\x3d','\x69','\x46','\x95','\x21',
	'\xdd','\x8c','\xf6','\x09','\x67','\x3d','\x6b','\x52','\xd2','\x54','\xe0','\x97','\x2b','\x0e','\x7f','\xb4',
	'\xbc','\xe6','\xb6','\xc5','\x13','\x98','\x30','\x27','\xfb','\x20','\xd9','\x75','\x80','\xba','\xdd','\x8b',
	'\xef','\xe3','\xa9','\x85','\xd3','\x56','\xe8','\xc0','\xf8','\x0f','\x86','\xd7','\x6a','\x4b','\xaf','\xa6',
	'\x76','\x85','\xd0','\x49','\xa6','\x79','\x94','\x1d','\xc8','\x23','\xe7','\xbd','\xe8','\xbf','\xf6','\x05',
	'\x50','\xca','\x2b','\x11','\x8e','\xff','\x34','\x3e','\x6d','\x5a','\xfc','\x26','\xf9','\x18','\xb0','\xa8',
	'\x7f','\xb3','\xba','\xdd','\x8a','\xe9','\xc8','\x23','\xe6','\xb6','\xc5','\x14','\x9f','\x54','\xde','\x8f',
};

/* ATTACK RATE e[u (ɗvNbN) */
const int sid_attack_table[16] = {
	   2000,	/* 0*/
	   8000,	/* 1*/
	  16000,	/* 2*/
	  24000,	/* 3*/
	  38000,	/* 4*/
	  56000,	/* 5*/
	  68000,	/* 6*/
	  80000,	/* 7*/
	 100000,	/* 8*/
	 250000,	/* 9*/
	 500000,	/*10*/
	 800000,	/*11*/
	1000000,	/*12*/
	3000000,	/*13*/
	5000000,	/*14*/
	8000000,	/*15*/
};

/* DECAY/RELEASE RATE e[u (ɗvNbN) */
const int sid_decay_release_table[16] = {
	    6000,	/* 0*/
	   24000,	/* 1*/
	   48000,	/* 2*/
	   72000,	/* 3*/
	  114000,	/* 4*/
	  168000,	/* 5*/
	  204000,	/* 6*/
	  240000,	/* 7*/
	  300000,	/* 8*/
	  750000,	/* 9*/
	 1500000,	/*10*/
	 2400000,	/*11*/
	 3000000,	/*12*/
	 9000000,	/*13*/
	15000000,	/*14*/
	24000000,	/*15*/
};

/****************************************************************************
 *
 ****************************************************************************/

void
sid_setup(SID* sid)
{
	SIDVOICE* voice;
	unsigned char* reg;
	int i_voice;

	/* VOICE1`3̃p[^ǂݍ݂܂B */
	voice = sid->voice;
	reg = sid->reg;
	for(i_voice = 0; i_voice < 3; i_voice++) {
		voice->to.freq	= reg[0] |  reg[1]        << 8;
		voice->wg.pw	= reg[2] | (reg[3] & 0xf) << 8;
		voice->wg.wf	= (reg[4] >> 4) & 0xf;
		voice->eg.atk	= (reg[5] >> 4) & 0xf;
		voice->eg.dcy	=  reg[5]       & 0xf;
		voice->eg.stn	= (reg[6] >> 4) & 0xf;
		voice->eg.rls	=  reg[6]       & 0xf;
		voice++;
		reg += 7; /* 7regs/voice */
	}

	/* ̑̃p[^ǂݍ݂܂B */
	sid->volume = sid->reg[24] & 0xf;
}

/****************************************************************************
 *	AvP[Vp֐
 ****************************************************************************/

void
sid_reset(SID* sid, int clock)
{
	memset(sid, 0, sizeof(SID));

	/* * 1TvɉNbNiނ?
	 *	1Tv̌oߎ = 1 / SPEAKER_FREQUENCY
	 *	1NbŇoߎ = 1 / clock
	 *	1TṽNbN = (1 / SPEAKER_FREQUENCY) / (1 / clock)
	 *				  = clock / SPEAKER_FREQUENCY
	 *	A[  = clock % SPEAKER_FREQUENCY
	 *	 clock(=1000000O) > SPEAKER_FREQUENCY(=16000) Ɖ肷B
	 */
	if(clock < SPEAKER_FREQUENCY) DIE();
	sid->clock.n = clock % SPEAKER_FREQUENCY;
	sid->clock.u = clock / SPEAKER_FREQUENCY;
}

unsigned char
sid_read(SID* sid, int regno)
{
	/* TODO: Ή */
	return -1;
}

void
sid_write(SID* sid, int regno, unsigned char data)
{
	int i_voice;
	SIDEG* eg;

	if(regno < 0 || 31 < regno) DIE();

	/* GATE ON/OFF fB */
	switch(regno) {
	case  4: i_voice = 0; break;
	case 11: i_voice = 1; break;
	case 18: i_voice = 2; break;
	default: goto NO_GATE;
	}
	if(!(sid->reg[regno] ^ data) & 1) goto NO_GATE; /* GATE bit ωȂ */
	eg = &sid->voice[i_voice].eg;
	if(data & 1) {
		eg->cycle = SIDEG_ATTACK;
	} else {
		if(eg->cycle <= SIDEG_RELEASE) goto NO_GATE; /* {0,RELEASE}GATE OFF͖ */
		eg->cycle = SIDEG_RELEASE;
	}
	eg->progress = 0;
NO_GATE:

	/* WX^li[B */
	sid->reg[regno] = data;
}

